/**
 * @file   backtrellis.c
 * @author Akinobu LEE
 * @date   Tue Feb 22 15:40:01 2005
 * 
 * <JA>
 * @brief  ñȥꥹ¸
 *
 * 1ѥη̤ñȥꥹȤ¸2ѥǻȤ뤿δؿ
 * ǤJulius Ǥϡ裱ѥõ˽üĤäƤñơ
 * λϽüե졼ࡤü٤ñȤȤ
 * ¸졤裲ѥǤν椫õԤޤ
 * 裱ѥǥե졼ऴȤ˻ĤñΤȤ֥ȥꥹñס
 * ȥꥹñνΤñȥꥹפȸƤӤޤ
 *
 * ȥꥹñϡ裱ѥǧ˳ƥե졼ऴȤ¸ޤ
 * 裱ѥλ塤ȥꥹΤ֤ȥե졼ऴȤΥǥå
 * ޤ
 *
 * 裲ѥǤϡñȥꥹ򻲾Ȥ
 * ƻ֡ϥե졼ˤˤŸǽʲΥꥹȤȤȤˡ
 * 裱ѥǤ()٤裲ѥˤ벾̤Ÿʬ
 * ꥹȤѤޤΤߤ顤ñȥꥹϡ֥Хåȥꥹ
 * ȤƤФƤޤ
 * </JA>
 * 
 * <EN>
 * @brief  Word trellis related operations
 *
 * Functions to store the result of the 1st pass as "word trellis", and
 * to access them from the 2nd pass is defined in this file.  On the
 * 1st pass of Julius, all the promising words whose word end has been
 * survived at the 1st pass will be stored as "word trellis", which
 * consists of trellis words that contains informations about the surviving
 * words: word boundary, accumulated score and word history.
 *
 * The trellis word will be stored per frame while recognition on the 1st pass.
 * After the 1st pass ended, the word trellis will be re-formed and
 * indexed by the frames to prepare for the access from the 2nd pass.
 *
 * In the 2nd pass of reverse stack decoding, this word trellis will be
 * used to constrain the word hypothesis, and also used to estimate
 * the score of unseen area by the obtained backward scores in the 1st pass.
 * Thus the word trellis information is also called as "back trellis" in
 * Julius.
 * </EN>
 * 
 * $Revision:$
 * 
 */
/*
 * Copyright (c) 1991-2005 Kawahara Lab., Kyoto University
 * Copyright (c) 2000-2005 Shikano Lab., Nara Institute of Science and Technology
 * Copyright (c) 2005      Julius project team, Nagoya Institute of Technology
 * All rights reserved
 */

#include <julius.h>

/** 
 * <JA>
 * ñȥꥹݻ ñȥꥹ ¤Τ(ư1¹)
 * 
 * @param bt [in]  ñȥꥹ ¤ΤؤΥݥ 
 * </JA>
 * <EN>
 * Initialize backtrellis that will hold the whole word trellis
 * (called once on startup).
 * 
 * @param bt [in] pointer to the backtrellis structure to initialize
 * </EN>
 */
void
bt_init(BACKTRELLIS *bt)
{
  bt->num  = NULL;
  bt->rw   = NULL;
  bt->list = NULL;
  bt->root = NULL;
}

/** 
 * <JA>
 * ǧѤ ñȥꥹ ¤Τ (ǧϻȤ˼¹)
 * 
 * @param bt [in] оݤȤñȥꥹ¤ΤؤΥݥ
 * </JA>
 * <EN>
 * Prepare backtrellis for the next input (called at beginning of each
 * speech segment).
 * 
 * @param bt [in] pointer to the word trellis structure
 * </EN>
 */
void
bt_prepare(BACKTRELLIS *bt)
{
  TRELLIS_ATOM *tre, *tmp;

  /* free previously allocated data */
  if (bt->root != NULL) free(bt->root);
  if (bt->rw != NULL) free(bt->rw);
  if (bt->num != NULL) free(bt->num);
  bt->num = NULL;
  bt->rw = NULL;
  bt->root = NULL;

  /* free actual trellis atoms previously allocated */
  tre = bt->list;
  while(tre) {
    tmp = tre->next;
    free(tre);
    tre = tmp;
  }
  /* reset entry point */
  bt->list = NULL;

}  



/** 
 * <JA>
 * 1ѥǽиȥꥹññ콪üΥȥꥹˤǼ롥
 *
 * ǤϳǼԤ1ѥλ bt_relocate_rw() 
 * ե졼˺֤롥
 * 
 * @param bt [i/o] ȥꥹñǼХåȥꥹ¤
 * @param tatom [in] иȥꥹñؤΥݥ
 * </JA>
 * <EN>
 * Store a trellis word generated on the 1st pass for the 2nd pass.
 *
 * This function just store the new atom into backtrellis.
 * They will be re-located per frame after 1st pass for quick access
 * in the 2nd pass.
 * 
 * @param bt [i/o] backtrellis structure to store the trellis word
 * @param tatom [in] the trellis word to be stored
 * </EN>
 */
void
bt_store(BACKTRELLIS *bt, TRELLIS_ATOM *tatom)
{
#ifdef WORD_GRAPH
  tatom->within_wordgraph = FALSE;
#endif
  tatom->next = bt->list;
  bt->list = tatom;
  /*j_printf("  %3d: %15s %f < %3d %15s\n", tatom->endtime,
	 winfo->woutput[tatom->wid], tatom->backscore,
	 tatom->begintime, winfo->woutput[tatom->last_tre->wid]);*/
}

/** 
 * <JA>
 * 1ѥλ, Ǽ줿ñȥꥹե졼˺֤롥
 * 
 * @param bt [i/o] ñȥꥹ¤
 * </JA>
 * <EN>
 * Re-locate the stored atom lists per frame (will be called after the
 * 1st pass).
 * 
 * @param bt [i/o] word trellis structure
 * </EN>
 */
void
bt_relocate_rw(BACKTRELLIS *bt)
{
  TRELLIS_ATOM *tre;
  int t;
  int totalnum, n;

  if (bt->framelen == 0) {
    bt->num = NULL;
    return;
  }

  bt->num = (int *)mymalloc(sizeof(int) * bt->framelen);

  /* count number of trellis atom (= survived word end) for each frame */
  for (t=0;t<bt->framelen;t++) bt->num[t] = 0;
  totalnum = 0;
  for (tre=bt->list;tre;tre=tre->next) {
#ifdef SP_BREAK_CURRENT_FRAME
    /* the last frame (when triggered from sp to non-sp) should be discarded */
    if (tre->endtime >= bt->framelen) continue;
#endif      
    bt->num[tre->endtime]++;
    totalnum++;
  }
  /* if no atom found, return here with all bt->num[t] set to 0 */
  if (totalnum <= 0) {
    free(bt->num);
    bt->num = NULL;
    return;
  }
  
  /* allocate area */
  bt->root = (TRELLIS_ATOM **)mymalloc(sizeof(TRELLIS_ATOM *) * totalnum);
  bt->rw  = (TRELLIS_ATOM ***)mymalloc(sizeof(TRELLIS_ATOM **) * bt->framelen);
  n = 0;
  for (t=0;t<bt->framelen;t++) {
    if (bt->num[t] > 0) {
      bt->rw[t] = (TRELLIS_ATOM **)&(bt->root[n]);
      n += bt->num[t];
    }
  }
  /* then store the atoms */
  for (t=0;t<bt->framelen;t++) bt->num[t] = 0;
  for (tre=bt->list;tre;tre=tre->next) {
#ifdef SP_BREAK_CURRENT_FRAME
    /* the last frame (when triggered from sp to non-sp) should be discarded */
    if (tre->endtime >= bt->framelen) continue;
#endif      
    t = tre->endtime;
    
    bt->rw[t][bt->num[t]] = tre;
    bt->num[t]++;
  }
}


/* ʲδؿ bt_relocate_rw ¹ԸˤΤ߻ѲǽȤʤ롥*/
/* functions below this line should be called after bt_relocate_rw() */

#ifdef SP_BREAK_CURRENT_FRAME

/** 
 * <JA>
 * ༡ǥǥ󥰻, 裱ѥλˡ
 * ϥȤξü˻Ĥäñ첾Ф, 
 * 2ѥˤ/ǽȤƥåȤ롥
 * 
 * @param bt [in] ñȥꥹ¤
 * </JA>
 * <EN>
 * When using progressive decoding with short pause segmentation,
 * This function extracts the best word hypothesis on head and tail of
 * the current input segment just after the 1st pass ends,
 * and store them as start/end word in the following 2nd pass.
 * 
 * @param bt [in] word trellis structure
 * </EN>
 */
void
set_terminal_words(BACKTRELLIS *bt)
{
  LOGPROB maxscore;
  int i,t;

  if (bt->num == NULL) return;

  maxscore = LOG_ZERO;
  /* find last frame where a word exists */
  for(t=bt->framelen-1;t>=0;t--) {
    if (bt->num[t] > 0) break;
  }
  /* get maximum word hypothesis at that frame */
  for(i=0;i<bt->num[t];i++) {
    if (maxscore < (bt->rw[t][i])->backscore) {
      maxscore = (bt->rw[t][i])->backscore;
      sp_break_2_begin_word = (bt->rw[t][i])->wid;
    }
  }
  maxscore = LOG_ZERO;
  /* find first frame where a word exists */
  for(t=0;t<bt->framelen;t++) {
    if (bt->num[t] > 0) break;
  }
  /* get maximum word hypothesis at that frame */
  for(i=0;i<bt->num[t];i++) {
    if (maxscore < (bt->rw[t][i])->backscore) {
      maxscore = (bt->rw[t][i])->backscore;
      sp_break_2_end_word = (bt->rw[t][i])->wid;
    }
  }
#ifdef SP_BREAK_DEBUG
  printf("2nd pass begin word: %s\n",
	 (sp_break_2_begin_word == WORD_INVALID) ? "WORD_INVALID" : winfo->wname[sp_break_2_begin_word]);
  printf("2nd pass end word: %s\n",
	 (sp_break_2_end_word == WORD_INVALID) ? "WORD_INVALID" : winfo->wname[sp_break_2_end_word]);
#endif
}
#endif /* SP_BREAK_CURRENT_FRAME */


/* the outprob on the trellis connection point should be discounted */
/** 
 * <JA>
 * 1ѥλ, 2ѥǤΥȥꥹ³׻Τˡ
 * ֤ϤäƳƥȥꥹñνüκǽ֤ν٤Ʒ׻,
 * Ѥ麹Ƥ裲ѥǤϡ³ˤ
 * ³θ³ξ֤٤Ʒ׻롥
 * 
 * @param wchmm [in] ڹ¤
 * @param bt [in] ñȥꥹ¤
 * @param param [in] ϥѥ᡼
 * </JA>
 * <EN>
 * Discount the output probabilities of the last state from the accumulated
 * score on word edge for all trellis words survived on the 1st pass,
 * for the acoustic re-computation on the 2nd pass.
 * The acousitic likelihood of the word edge state will be re-computed
 * when the next word hypotheses are expanded on the next 2nd pass.
 * 
 * @param wchmm [in] tree lexicon
 * @param bt [in] word trellis structure
 * @param param [in] input parameter
 * </EN>
 */
void
bt_discount_pescore(WCHMM_INFO *wchmm, BACKTRELLIS *bt, HTK_Param *param)
{
  int t,i;
  TRELLIS_ATOM *tre;

  if (bt->num == NULL) return;
  
  for (t=0; t<bt->framelen; t++) {
    for (i=0; i<bt->num[t]; i++) {
      tre = bt->rw[t][i];
#ifndef MULTIPATH_VERSION
      /* On normal version, both language score and the output prob. score
	 at the connection point should removed on the trellis for the
	 later connection.  On multi-path mode, removing only the language
	 score is enough. */
      tre->backscore -= outprob_style(wchmm, wchmm->wordend[tre->wid], tre->last_tre->wid, t, param);
#endif /* MULTIPATH_VERSION */
#ifdef USE_NGRAM
#ifdef LM_FIX_DOUBLE_SCORING
      /* the LM score of the last word should be subtracted, because
	 their LM will be assigned by 3-gram on the 2nd pass.
      */
      tre->backscore -= tre->lscore;
#endif
#endif
    }
  }

}

/** 
 * <JA>
 * bt_sort_rw()ѤqsortХå
 * 
 * @param a [in] ǣ
 * @param b [in] ǣ
 * 
 * @return 祽Ȥɬפ
 * </JA>
 * <EN>
 * qsort callback for bt_sort_rw().
 * 
 * @param a [in] first element
 * @param b [in] second element
 * 
 * @return a value needed to do upward sort.
 * </EN>
 */
static int
compare_wid(TRELLIS_ATOM **a, TRELLIS_ATOM **b)
{
  if ((*a)->wid > (*b)->wid) return 1;
  if ((*a)->wid < (*b)->wid) return -1;
  return 0;
}

/** 
 * <JA>
 * bt_relocate_rw() λ, ®Τ
 * Хåȥꥹ¤Υȥꥹñե졼ऴȤ
 * ñIDǥȤƤ
 * 
 * @param bt [i/o] ñȥꥹ¤
 * 
 * </JA>
 * <EN>
 * Sort the trellis words in the backtrellis by the word IDs per each frame,
 * for rapid access on the 2nd pass.  This should be called just after
 * bt_relocate_rw() was called.
 * 
 * @param bt [i/o] word trellis structure
 * 
 * </EN>
 */
void
bt_sort_rw(BACKTRELLIS *bt)
{
  int t;

  if (bt->num == NULL) return;

  for (t=0;t<bt->framelen;t++) {
    qsort(bt->rw[t], bt->num[t], sizeof(TRELLIS_ATOM *),
	  (int (*)(const void *,const void *))compare_wid);
  }
}


/* ʲδؿϻbt_sort_rw() ƤФƤ뤳(2ѥ) */
/* functions below should be called after bt_sort_rw() */

/** 
 * <JA>
 * ñȥꥹλե졼ˡñνü뤫ɤ
 * 롥
 * 
 * @param bt [in] ñȥꥹ¤
 * @param t [in] 뽪üʥե졼
 * @param wkey [in] ññɣ
 * 
 * @return Ĥä礽ΥȥꥹñؤΥݥ󥿡Ĥʤ NULL
 * </JA>
 * <EN>
 * Search a word on the specified frame in a word trellis data.
 * 
 * @param bt [in] word trellis structure
 * @param t [in] word end frame on which to search
 * @param wkey [in] word ID to search
 * 
 * @return pointer to the found trellis word, or NULL if not found.
 * </EN>
 */
TRELLIS_ATOM *
bt_binsearch_atom(BACKTRELLIS *bt, int t, WORD_ID wkey)
{
  /* do binary search */
  /* assume rw are ordered by wid */
  int left, right, mid;
  TRELLIS_ATOM *tmp;
#ifdef WPAIR
  int i;
  LOGPROB maxscore;
  TRELLIS_ATOM *maxtre;
#endif
  
  if (bt->num[t] == 0) return(NULL);
  
  left = 0;
  right = bt->num[t] - 1;
  while (left < right) {
    mid = (left + right) / 2;
    if ((bt->rw[t][mid])->wid < wkey) {
      left = mid + 1;
    } else {
      right = mid;
    }
  }
  tmp = bt->rw[t][left];
  if (tmp->wid == wkey) {
#ifdef WPAIR
    /* same word with different context will be found:
       most likely one will be returned */
    maxscore = LOG_ZERO;
    i = left;
    while (i >= 0 && (tmp = bt->rw[t][i])->wid == wkey) {
      if (maxscore < tmp->backscore) {
	maxscore = tmp->backscore;
	maxtre = tmp;
      }
      i--;
    }
    i = left;
    while (i < bt->num[t] && (tmp = bt->rw[t][i])->wid == wkey) {
      if (maxscore < tmp->backscore) {
	maxscore = tmp->backscore;
	maxtre = tmp;
      }
      i++;
    }

    tmp = maxtre;
#endif /* WORD_GRAPH */

    return(tmp);
  } else {
    return(NULL);
  }
}
